#include <iomanip>
#include <time.h>
#include <ctype.h>
#include <unistd.h>
#include <math.h>
#include <stdlib.h>     // For (s)random

#include <sstream>
#include <set>

#include <id3/misc_support.h>

#include "strmanip.h"
#include "imms.h"

string MyID3Tag::get_artist()
{
    char buffer[1024];
    ID3_Frame *myFrame = Find(ID3FID_LEADARTIST);
    if (myFrame)
    {
        myFrame->Field(ID3FN_TEXT).Get(buffer, 1024);
        return string_normalize(buffer);
    }
    return "";
}

string MyID3Tag::get_title()
{
    char buffer[1024];
    ID3_Frame *myFrame = Find(ID3FID_TITLE);
    if (myFrame)
    {
        myFrame->Field(ID3FN_TEXT).Get(buffer, 1024);
        return string_normalize(buffer);
    }
    return "";
}

Imms::Imms() : threshold(INITIAL_RATING), timehold(0),
    last_skipped(true), last_jumped(false), status(dont_rate)
{
    srandom(time(NULL));

    fout.open("/dev/null", ofstream::out | ofstream::app);
}

void Imms::setup(const char* _email, const char* _logfile)
{
    fout.close();
    fout.open(_logfile, ofstream::out | ofstream::app);
    email = _email;
    time_t t = time(NULL);
    fout << endl << endl << ctime(&t) << "User: '" << email << "'";
}

Status Imms::song_changed(const string &path, int time_left, bool jumped)
{
    if (access(path.c_str(), R_OK))
    {
        return status = dont_rate;
    }

    /*
     * Adjust rating on the last song
     */ 
    if (status == play_ok)
    {
        int mod;

        if (time_left > 2)
        {
            if (last_jumped && !jumped)
                mod = JUMPED_TO_SKIPPED;
            else if (jumped)
                mod = JUMPED_FROM;
            else if (last_skipped)
                mod = CONS_SKIP_RATE;
            else
                mod = FIRST_SKIP_RATE;
            last_skipped = true;
        }
        else
        {
            if (last_jumped)
                mod = JUMPED_TO_FINNISHED;
            else if (last_skipped)
                mod = FIRST_NON_SKIP_RATE;
            else
                mod = CONS_NON_SKIP_RATE;
            last_skipped = jumped = false;
        }

        fout << (!last_jumped && last_jumped ? "[Jump: To] " : "");
        fout << (jumped ? "[Jump: From] " : "");
        fout << (last_skipped && !jumped ? "[Skip: Manual] " : "");
        fout << "[Delta: " << setiosflags(ios::showpos) << mod <<
            resetiosflags (ios::showpos) << "] ";

        last_jumped = jumped;
    
        set_popm(mod);
    }
    else
    {
        last_jumped = jumped = false;
        status = play_ok;
    }

    string ext = "";
    string oldpath = id3tag.GetFileName() ? id3tag.GetFileName() : "";
    if (oldpath.length() > 3)
    {
        ext = string_tolower(oldpath.substr(oldpath.length() - 3, 3));
    }

    if (oldpath != "")
    {
        if (ext != "mp3" && ext != "ogg")
            fout << "[BadExt] ";
        else if (access(oldpath.c_str(), W_OK))
            fout << "[RO] ";
        else 
            id3tag.Update(ID3TT_ID3V2);
    }

    /*
     * Start processing the new song
     */ 

    id3tag.Clear();

    fout << endl << string(TERM_WIDTH - 20, '-') << endl;

    if (path.length() > TERM_WIDTH - 8)
        fout << "[Path: ..." << path.substr(path.length() - TERM_WIDTH + 11,
                TERM_WIDTH - 11) << "]" << endl;
    else
        fout << "[Path: " << path << "]" << endl;

    id3tag.Link(path.c_str());

    int popm = get_popm();

    fout << "[Rating: " << popm << "/" << setprecision(3) <<
        threshold << setprecision(5) << "/";

    if (!jumped && popm < threshold)
    {
        // Figure out how far below the threshold this song falls
        float total = pow(2, (threshold - popm)/EXPONENT_STEP) + BIAS;
        if (random() % 100 * total / 99 > 1) 
        {
            status = do_skip;
        }
        int twodigit = (int)(100 / total);
        fout << (float)twodigit / 100 << "] ";
    }
    else
        fout << "-] ";

    if (popm > threshold)
        threshold = threshold * (1 - THRESHOLD_GROW) + popm * THRESHOLD_GROW;
    else 
        threshold = threshold * (1 - THRESHOLD_DECAY) + popm * THRESHOLD_DECAY;

    /*
     * Try really hard to guess what this song is
     */ 

    string artist = id3tag.get_artist();
    string title = id3tag.get_title();

    bool artist_confirmed = false, title_confirmed = false;

    // Check if the artist/title specified in the tag are known
    if (artist != "")
        artist_confirmed = sqldb.check_artist(artist);

    if (artist_confirmed && title != "")
        title_confirmed = sqldb.check_title(artist, title);

#ifdef DEBUG
    cerr << "========================================" << endl;
    cerr << "path:\t" << path << endl;
    cerr << "artist:\t" << artist << (artist_confirmed ? " *" : "") << endl;
    cerr << "title:\t" << title << (title_confirmed ? " *" : "") << endl;
#endif

    while (!(artist_confirmed && title_confirmed))
    {
        int last_slash = path.find_last_of("/");
        string only_path = path.substr(0, last_slash);
        string filename = path.substr(last_slash + 1,
                path.length() - last_slash - 5);

        set<string> file_parts;
        imms_magic_parse(file_parts, filename);

        if (!artist_confirmed)
        {
            set<string> path_parts(file_parts);
            imms_magic_parse(path_parts, only_path);

#ifdef DEBUG
            cerr << "Path parts: " << endl;
            for (set<string>::iterator i = path_parts.begin();
                    i != path_parts.end(); i++)
            {
                cerr << " >> " << *i << endl;
            }
#endif

            // See if we can recognize any of the path parts
            for (set<string>::iterator i = path_parts.begin();
                    !artist_confirmed && i != path_parts.end(); i++)
            {
                if (artist_confirmed = sqldb.check_artist(*i))
                {
                    artist = *i;
                }
            }

            // If that failed, maybe this artist is new.
            // Try matching tag against the parts of the path to make sure
            if (!artist_confirmed && artist != "")
            {
                artist_confirmed =
                    (path_parts.find(artist) != path_parts.end());
            }

            // If we still don't know what to do - give up
            if (!artist_confirmed)
                break;
        }
        // If we know the title and were only missing the artist
        if (title != "")
            title_confirmed = sqldb.check_title(artist, title);
 
        // Can recognize the title by matching filename parts to the database?
        for (set<string>::iterator i = file_parts.begin();
                !title_confirmed && i != file_parts.end(); i++)
        {
            if (title_confirmed = sqldb.check_title(artist, *i))
            {
                title = *i;
            }
            else
            {
                title_confirmed = (i->find(title) != string::npos &&
                        title.length() > 2);
            }
        }
    
        if (title_confirmed)
            break;

        // If not, fall back on path comparisons
        if (title != "")
        {
            if (title_confirmed = (file_parts.find(title) != file_parts.end()))
                break;
        }

        file_parts.erase(artist);
        if (file_parts.size() == 1)
        {
            title = *file_parts.begin();   
            title_confirmed = true;
        }

        break;
    }

    bool identified = artist_confirmed && title_confirmed;

    time_t last_time = identified ? sqldb.get_last(artist, title)
        : sqldb.get_last(path);

    last_time = time(NULL) - last_time;
    last_time /= 3600;
    if (last_time > 10000)
        last_time = 10000;

    fout << "[Time: ";
    if  (last_time == 10000)
        fout << "Never";
    else
        fout << last_time;

    fout << "/" << setprecision(3) << timehold << setprecision(5) << "] ";

    if (status == play_ok && !jumped && last_time < timehold)
        status = jump_random;

    if (status == play_ok)
    {
        if (identified)
            sqldb.update_last(artist, title, time(NULL));
        else
            sqldb.update_last(path, time(NULL));
    }

    if (last_time != 10000)
        timehold = timehold * (1 - TIMEHOLD_PART) + last_time * TIMEHOLD_PART;

    if (!identified) 
        fout << "[Unknown] ";

    if (status == do_skip)
        fout << "[Skip: Auto] ";
    else if (status == jump_random)
        fout << "[Skip: Time] ";

    fout.flush();

    return status;
}

void Imms::set_popm(int mod)
{
    ID3_Frame* popmframe = id3tag.Find(ID3FID_POPULARIMETER,
            ID3FN_EMAIL, email.c_str());

    if (!popmframe)
    {
        fout << "[New] "; 
        popmframe = new ID3_Frame(ID3FID_POPULARIMETER);
        popmframe->GetField(ID3FN_EMAIL)->Set(email.c_str()); 
        id3tag.AttachFrame(popmframe); // id3 takes ownership
        popmframe = id3tag.Find(ID3FID_POPULARIMETER,
                ID3FN_EMAIL, email.c_str());
    }

    ID3_Field* ratingfield = popmframe->GetField(ID3FN_RATING);

    int newrating = get_popm() + mod;

    if (newrating > 150)
        newrating = 150;

    if (newrating < 50)
        newrating = 50;

    ratingfield->Set(newrating);
}

int Imms::get_popm()
{
    ID3_Frame* popmframe = id3tag.Find(ID3FID_POPULARIMETER,
            ID3FN_EMAIL, email.c_str());

    if (popmframe)
    {
        ID3_Field* ratingfield = popmframe->GetField(ID3FN_RATING);
        if (ratingfield && ratingfield->Get() > 0)
        {
            return ratingfield->Get();
        }
    }
    return INITIAL_RATING;
}
